/*
 * lid.cxx
 *
 * Line Interface Device
 *
 * Open Phone Abstraction Library
 *
 * Copyright (c) 1999-2000 Equivalence Pty. Ltd.
 *
 * The contents of this file are subject to the Mozilla Public License
 * Version 1.0 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS"
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See
 * the License for the specific language governing rights and limitations
 * under the License.
 *
 * The Original Code is Open H323 Library.
 *
 * The Initial Developer of the Original Code is Equivalence Pty. Ltd.
 *
 * Contributor(s): ______________________________________.
 *
 * $Log: lid.cxx,v $
 * Revision 1.32  2000/07/13 16:03:25  robertj
 * Removed transmission of 1 byte repeat CNG frames in G.723.1 as it crashes other peoples stacks.
 *
 * Revision 1.31  2000/07/12 10:25:37  robertj
 * Renamed all codecs so obvious whether software or hardware.
 *
 * Revision 1.30  2000/07/09 15:23:00  robertj
 * Changed G.728 not to use Cisco hack. Cisco is just wrong!
 * Fixed output of silence in G.711 so works with any sized frame.
 *
 * Revision 1.29  2000/07/02 14:09:49  craigs
 * Fill uLaw and aLaw silence with 0x80 rather than 0x00
 *
 * Revision 1.28  2000/06/19 00:32:22  robertj
 * Changed functionf or adding all lid capabilities to not assume it is to an endpoint.
 *
 * Revision 1.27  2000/06/01 07:52:30  robertj
 * Changed some LID capability code back again so does not unneedfully break existing API.
 *
 * Revision 1.26  2000/05/30 10:19:28  robertj
 * Added function to add capabilities given a LID.
 * Improved LID capabilities so cannot create one that is not explicitly supported.
 *
 * Revision 1.25  2000/05/24 06:43:16  craigs
 * Added routines to get xJack volume
 * Fixed problem with receiving G>723.1 NULL frames
 *
 * Revision 1.24  2000/05/11 03:47:48  craigs
 * Added extra debugging
 *
 * Revision 1.23  2000/05/10 04:05:34  robertj
 * Changed capabilities so has a function to get name of codec, instead of relying on PrintOn.
 *
 * Revision 1.22  2000/05/04 12:56:43  robertj
 * Fixed GNU warning.
 *
 * Revision 1.21  2000/05/04 11:52:35  robertj
 * Added Packets Too Late statistics, requiring major rearrangement of jitter
 *    buffer code, not also changes semantics of codec Write() function slightly.
 *
 * Revision 1.20  2000/05/02 04:32:27  robertj
 * Fixed copyright notice comment.
 *
 * Revision 1.19  2000/04/30 03:57:14  robertj
 * Added PTRACE of read/write frame sizes required by LID.
 *
 * Revision 1.18  2000/04/19 02:04:30  robertj
 * BeOS port changes.
 *
 * Revision 1.17  2000/04/14 17:18:07  robertj
 * Fixed problem with error reporting from LID hardware.
 *
 * Revision 1.16  2000/04/10 17:45:11  robertj
 * Added higher level "DialOut" function for PSTN lines.
 * Added hook flash function.
 *
 * Revision 1.15  2000/04/05 18:04:12  robertj
 * Changed caller ID code for better portability.
 *
 * Revision 1.14  2000/04/03 19:25:14  robertj
 * Optimised G.711 codec to read/write larger chunks of data.
 *
 * Revision 1.13  2000/03/31 19:50:51  robertj
 * Fixed receiver loop being able to deal with RTP packets smaller than expected.
 *
 * Revision 1.12  2000/03/30 19:32:35  robertj
 * Added swab function which seems to be missing on Linux.
 *
 * Revision 1.11  2000/03/30 01:57:16  robertj
 * Added hacks so G.728 works with (I think) broken cisco gateways.
 *
 * Revision 1.10  2000/03/29 21:01:52  robertj
 * Changed codec to use number of frames rather than number of bytes.
 * Added function on LID to get available codecs.
 * Fixed codec table for G.729 codec
 *
 * Revision 1.9  2000/03/28 05:22:05  robertj
 * Fixed translation of text country code to numeric code.
 *
 * Revision 1.8  2000/03/23 23:36:49  robertj
 * Added more calling tone detection functionality.
 *
 * Revision 1.7  2000/03/21 03:06:50  robertj
 * Changes to make RTP TX of exact numbers of frames in some codecs.
 *
 * Revision 1.6  2000/01/13 12:39:29  robertj
 * Added string based country codes to LID.
 *
 * Revision 1.5  2000/01/13 04:03:46  robertj
 * Added video transmission
 *
 * Revision 1.4  2000/01/07 10:01:26  robertj
 * GCC/Linux compatibility
 *
 * Revision 1.3  2000/01/07 08:28:09  robertj
 * Additions and changes to line interface device base class.
 *
 * Revision 1.2  1999/12/29 01:18:07  craigs
 * Fixed problem with codecs other than G.711 not working after reorganisation
 *
 * Revision 1.1  1999/12/23 23:02:36  robertj
 * File reorganision for separating RTP from H.323 and creation of LID for VPB support.
 *
 */

#include <ptlib.h>
#include "lid.h"
#include "h323.h"
#include "h245.h"


#define new PNEW

//#define CISCO_G728_HACK

#if defined(CISCO_G728_HACK) && (defined(P_LINUX) || defined(__BEOS__))
void swab(char * from, char * to, size_t len)
{
  while (len > 1) {
    char b = from[0];
    to[0] = from[1];
    to[1] = b;
    from += 2;
    to += 2;
    len -= 2;
  }
}
#endif


///////////////////////////////////////////////////////////////////////////////

#if PTRACING
const char * const OpalLineInterfaceDevice::CallProgressTonesNames[NumTones] = {
  "DialTone", "RingTone", "BusyTone", "ClearTone"
};
#endif


OpalLineInterfaceDevice::OpalLineInterfaceDevice()
{
  os_handle = -1;
  osError = 0;
  countryCode = UnknownCountry;
}


BOOL OpalLineInterfaceDevice::IsOpen() const
{
  return os_handle >= 0;
}


BOOL OpalLineInterfaceDevice::Close()
{
  if (os_handle < 0)
    return FALSE;

  os_handle = -1;
  return TRUE;
}


BOOL OpalLineInterfaceDevice::IsLineTerminal(unsigned)
{
  return FALSE;
}


BOOL OpalLineInterfaceDevice::IsLinePresent(unsigned)
{
  return TRUE;
}


BOOL OpalLineInterfaceDevice::HookFlash(unsigned line, unsigned flashTime)
{
  if (!IsLineOffHook(line))
    return FALSE;

  if (!SetLineOnHook(line))
    return FALSE;

  PThread::Current()->Sleep(flashTime);

  return SetLineOffHook(line);
}


BOOL OpalLineInterfaceDevice::IsLineRinging(unsigned, DWORD *)
{
  return FALSE;
}


BOOL OpalLineInterfaceDevice::RingLine(unsigned, DWORD)
{
  return FALSE;
}


BOOL OpalLineInterfaceDevice::IsLineDisconnected(unsigned line)
{
  return IsToneDetected(line) == BusyTone;
}


BOOL OpalLineInterfaceDevice::SetLineToLineDirect(unsigned, unsigned, BOOL)
{
  return FALSE;
}


BOOL OpalLineInterfaceDevice::IsLineToLineDirect(unsigned, unsigned)
{
  return FALSE;
}


PIntArray OpalLineInterfaceDevice::GetPayloadTypes() const
{
  PIntArray codecs(1);
  codecs[0] = RTP_DataFrame::L16_Mono;
  return codecs;
}


BOOL OpalLineInterfaceDevice::SetReadCodec(unsigned, RTP_DataFrame::PayloadTypes)
{
  return FALSE;
}


BOOL OpalLineInterfaceDevice::SetWriteCodec(unsigned, RTP_DataFrame::PayloadTypes)
{
  return FALSE;
}


BOOL OpalLineInterfaceDevice::StopReadCodec(unsigned)
{
  return FALSE;
}


BOOL OpalLineInterfaceDevice::StopWriteCodec(unsigned)
{
  return FALSE;
}


BOOL OpalLineInterfaceDevice::SetReadFrameSize(unsigned, PINDEX)
{
  return FALSE;
}


PINDEX OpalLineInterfaceDevice::SetWriteFrameSize(unsigned, PINDEX)
{
  return FALSE;
}


unsigned OpalLineInterfaceDevice::GetAverageSignalLevel(unsigned)
{
  return UINT_MAX;
}


BOOL OpalLineInterfaceDevice::EnableAudio(unsigned, BOOL)
{
  return TRUE;
}


BOOL OpalLineInterfaceDevice::SetRecordVolume(unsigned, unsigned)
{
  return FALSE;
}

BOOL OpalLineInterfaceDevice::GetRecordVolume(unsigned, unsigned &)
{
  return FALSE;
}


BOOL OpalLineInterfaceDevice::SetPlayVolume(unsigned, unsigned)
{
  return FALSE;
}

BOOL OpalLineInterfaceDevice::GetPlayVolume(unsigned, unsigned &)
{
  return FALSE;
}


OpalLineInterfaceDevice::AECLevels OpalLineInterfaceDevice::GetAEC(unsigned)
{
  return AECError;
}


BOOL OpalLineInterfaceDevice::SetAEC(unsigned, AECLevels)
{
  return FALSE;
}


BOOL OpalLineInterfaceDevice::GetCallerID(unsigned, PString & id, BOOL)
{
  id = PString();
  return FALSE;
}


BOOL OpalLineInterfaceDevice::SetCallerID(unsigned, const PString &)
{
  return FALSE;
}


BOOL OpalLineInterfaceDevice::PlayDTMF(unsigned, const char *, DWORD, DWORD)
{
  return FALSE;
}


char OpalLineInterfaceDevice::ReadDTMF(unsigned)
{
  return '\0';
}


BOOL OpalLineInterfaceDevice::GetRemoveDTMF(unsigned)
{
  return FALSE;
}


BOOL OpalLineInterfaceDevice::SetRemoveDTMF(unsigned, BOOL)
{
  return FALSE;
}


OpalLineInterfaceDevice::CallProgressTones
                            OpalLineInterfaceDevice::IsToneDetected(unsigned)
{
  return NumTones;
}


OpalLineInterfaceDevice::CallProgressTones
               OpalLineInterfaceDevice::WaitForToneDetect(unsigned line, unsigned timeout)
{
  static const unsigned sampleRate = 25;

  timeout = (timeout+sampleRate-1)/sampleRate;

  unsigned retry;
  for (retry = 0; retry < timeout; retry++) {
    CallProgressTones tone = IsToneDetected(line);
    if (tone != NumTones) {
      PTRACE(2, "LID\tTone " << tone << " detected after " << (retry*sampleRate) << " ms");
      return tone;
    }

    PThread::Current()->Sleep(25);
  }

  PTRACE(3, "LID\tTone detection timeout " << (retry*sampleRate) << " ms");
  return NumTones;
}


BOOL OpalLineInterfaceDevice::WaitForTone(unsigned line,
                                          CallProgressTones tone,
                                          unsigned timeout)
{
  return WaitForToneDetect(line, timeout) == tone;
}


BOOL OpalLineInterfaceDevice::PlayTone(unsigned, CallProgressTones)
{
  return FALSE;
}


BOOL OpalLineInterfaceDevice::IsTonePlaying(unsigned)
{
  return FALSE;
}


BOOL OpalLineInterfaceDevice::StopTone(unsigned)
{
  return FALSE;
}


OpalLineInterfaceDevice::CallProgressTones
                OpalLineInterfaceDevice::DialOut(unsigned line,
                                                 const PString & number,
                                                 BOOL requireTone)
{
  PTRACE(3, "LID\tDialOut to " << number);

  if (IsLineTerminal(line))
    return NumTones;

  if (!SetLineOffHook(line))
    return NumTones;

  // Should get dial tone within 2 seconds of going off hook
  if (!WaitForTone(line, DialTone, 2000)) {
    if (requireTone)
      return DialTone;
  }

  // Dial the string
  PINDEX lastPos = 0;
  PINDEX nextPos;
  while ((nextPos = number.FindOneOf("!@,")) != P_MAX_INDEX) {
    PlayDTMF(line, number(lastPos, nextPos-1));
    lastPos = nextPos+1;
    switch (number[nextPos]) {
      case '!' :
        if (!HookFlash(line))
          return NumTones;
        break;

      case '@' :
        if (!WaitForTone(line, DialTone, 3000)) {
          if (requireTone)
            return DialTone;
        }
        break;

      case ',' :
        PThread::Current()->Sleep(2000);
        break;
    }
  }

  PlayDTMF(line, number.Mid(lastPos));

  // Wait for busy or ring back
  CallProgressTones tone;
  while ((tone = WaitForToneDetect(line, 5000)) != NumTones) {
    if (tone != DialTone)
      return tone;
  }

  if (requireTone)
    return NumTones;

  return RingTone;
}


static const char * CountryNames[OpalLineInterfaceDevice::NumCountryCodes] = {
  "Japan", "Albania", "Algeria", "AmericanSamoa", "Germany", "Anguilla", "AntiguaAndBarbuda",
  "Argentina", "Ascension", "Australia", "Austria", "Bahamas", "Bahrain", "Bangladesh",
  "Barbados", "Belgium", "Belize", "Benin", "Bermudas", "Bhutan", "Bolivia", "Botswana",
  "Brazil", "BritishAntarcticTerritory", "BritishIndianOceanTerritory",
  "BritishVirginIslands", "BruneiDarussalam", "Bulgaria", "Myanmar", "Burundi",
  "Byelorussia", "Cameroon", "Canada", "CapeVerde", "CaymanIslands",
  "CentralAfricanRepublic", "Chad", "Chile", "China", "Colombia", "Comoros", "Congo",
  "CookIslands", "CostaRica", "Cuba", "Cyprus", "Czechoslovakia", "Cambodia",
  "DemocraticPeoplesRepublicOfKorea", "Denmark", "Djibouti", "DominicanRepublic",
  "Dominica", "Ecuador", "Egypt", "ElSalvador", "EquatorialGuinea", "Ethiopia",
  "FalklandIslands", "Fiji", "Finland", "France", "FrenchPolynesia",
  "FrenchSouthernAndAntarcticLands", "Gabon", "Gambia", "Germany2", "Angola", "Ghana",
  "Gibraltar", "Greece", "Grenada", "Guam", "Guatemala", "Guernsey", "Guinea", "GuineaBissau",
  "Guayana", "Haiti", "Honduras", "Hongkong", "Hungary", "Iceland", "India", "Indonesia",
  "Iran", "Iraq", "Ireland", "Israel", "Italy", "CotedIvoire", "Jamaica", "Afghanistan",
  "Jersey", "Jordan", "Kenya", "Kiribati", "KoreaRepublic", "Kuwait", "Lao", "Lebanon",
  "Lesotho", "Liberia", "Libya", "Liechtenstein", "Luxemborg", "Macao", "Madagascar",
  "Malaysia", "Malawi", "Maldives", "Mali", "Malta", "Mauritania", "Mauritius", "Mexico",
  "Monaco", "Mongolia", "Montserrat", "Morocco", "Mozambique", "Nauru", "Nepal",
  "Netherlands", "NetherlandsAntilles", "NewCaledonia", "NewZealand", "Nicaragua",
  "Niger", "Nigeria", "Norway", "Oman", "Pakistan", "Panama", "PapuaNewGuinea", "Paraguay",
  "Peru", "Philippines", "Poland", "Portugal", "PuertoRico", "Qatar", "Romania", "Rwanda",
  "SaintKittsAndNevis", "SaintCroix", "SaintHelenaAndAscension", "SaintLucia",
  "SanMarino", "SaintThomas", "SaoTomeAndPrincipe", "SaintVicentAndTheGrenadines",
  "SaudiArabia", "Senegal", "Seychelles", "SierraLeone", "Singapore", "SolomonIslands",
  "Somalia", "SouthAfrica", "Spain", "SriLanka", "Sudan", "Suriname", "Swaziland", "Sweden",
  "Switzerland", "Syria", "Tanzania", "Thailand", "Togo", "Tonga", "TrinidadAndTobago",
  "Tunisia", "Turkey", "TurksAndCaicosIslands", "Tuvalu", "Uganda", "Ukraine",
  "UnitedArabEmirates", "UnitedKingdom", "UnitedStates", "BurkinaFaso", "Uruguay",
  "USSR", "Vanuatu", "VaticanCityState", "Venezuela", "VietNam", "WallisAndFutuna",
  "WesternSamoa", "Yemen", "Yemen2", "Yugoslavia", "Zaire", "Zambia", "Zimbabwe"
};

PString OpalLineInterfaceDevice::GetCountryCodeName() const
{
  if (countryCode >= 0 && countryCode < NumCountryCodes)
    return CountryNames[countryCode];

  return "<Unknown>";
}


BOOL OpalLineInterfaceDevice::SetCountryCode(T35CountryCodes country)
{
  countryCode = country;
  PTRACE(2, "LID\tCountry set to " << GetCountryCodeName());
  return TRUE;
}


BOOL OpalLineInterfaceDevice::SetCountryCode(const PString & countryName)
{
  PINDEX code = Japan;
  while (code < NumCountryCodes) {
    if (countryName *= CountryNames[code])
      return SetCountryCode((T35CountryCodes)code);
    code++;
  }

  SetCountryCode(UnknownCountry);
  return FALSE;
}


PString OpalLineInterfaceDevice::GetErrorText() const
{
  return PChannel::GetErrorText(PChannel::Miscellaneous, osError);
}


void OpalLineInterfaceDevice::PrintOn(ostream & strm) const
{
  strm << GetName();
}


/////////////////////////////////////////////////////////////////////////////

OpalLineChannel::OpalLineChannel(OpalLineInterfaceDevice & dev, unsigned line, BOOL rd)
  : device(dev)
{
  lineNumber = line;
  reading = rd;
  codec = RTP_DataFrame::L16_Mono;

  if (reading) {
    device.SetReadCodec(line, codec);
    deblockingBuffer.SetSize(device.GetReadFrameSize(lineNumber));
    deblockingOffset = P_MAX_INDEX;
  }
  else {
    device.SetWriteCodec(line, codec);
    deblockingBuffer.SetSize(device.GetWriteFrameSize(lineNumber));
    deblockingOffset = 0;
  }
}


OpalLineChannel::~OpalLineChannel()
{
  Close();
}


PString OpalLineChannel::GetName() const
{
  return device.GetName() + psprintf("-%u", lineNumber);
}


BOOL OpalLineChannel::Close()
{
  if (reading)
    return device.StopReadCodec(lineNumber);
  else
    return device.StopWriteCodec(lineNumber);
}


BOOL OpalLineChannel::SetCodec(RTP_DataFrame::PayloadTypes theCodec)
{
  codec = theCodec;

  if (reading) {
    if (!device.SetReadCodec(lineNumber, codec))
      return FALSE;
  }
  else {
    if (!device.SetWriteCodec(lineNumber, codec))
      return FALSE;
  }

  PTRACE(3, "LID\tCodec set to " << theCodec << ", frame size="
         << device.GetReadFrameSize(lineNumber) << '/'
         << device.GetWriteFrameSize(lineNumber));
  return TRUE;
}


BOOL OpalLineChannel::ReadFrame(BYTE * buffer)
{
  if (device.ReadFrame(lineNumber, buffer))
    return TRUE;

#if PTRACING
  if (device.GetErrorNumber() != 0)
    PTRACE(1, "LID\tDevice read frame error: " << device.GetErrorText());
#endif
  return FALSE;
}


BOOL OpalLineChannel::Read(void * buffer, PINDEX length)
{
  if (!reading) {
    lastError = Miscellaneous;
    osError = EINVAL;
    return FALSE;
  }

  BYTE * bufferPtr = (BYTE *)buffer;

  // Pick out special cases for G.723.1 based codecs (variable length frames)
  if (codec == RTP_DataFrame::G7231) {
    if (!ReadFrame(bufferPtr))
      return FALSE;

    switch (bufferPtr[0]&3) {
      case 0 : // 6.3kbps rate frame
        lastReadCount = 24;
        break;
      case 1 : // 5.3kbps rate frame
        lastReadCount = 20;
        break;
      case 2 : // a Silence Insertion Descriptor
        lastReadCount = 4;
        break;
      case 3 : // repeat last CNG frame
        lastReadCount = 0;
        break;
    }

    return TRUE;
  }

  // Are reblocking the hardware frame sizes to those expected by the RTP packets.
  lastReadCount = 0;
  device.SetReadFrameSize(lineNumber, length);
  PINDEX frameSize = device.GetReadFrameSize(lineNumber);

  while (length > 0) {
    if (deblockingOffset < frameSize) {
      PINDEX left = frameSize - deblockingOffset;
      if (left > length)
        left = length;
      memcpy(bufferPtr, &deblockingBuffer[deblockingOffset], left);
      lastReadCount += left;
      deblockingOffset += left;
      bufferPtr += left;
      length -= left;
    }
    else if (length < frameSize) {
      BYTE * deblockPtr = deblockingBuffer.GetPointer(frameSize);
      if (!ReadFrame(deblockPtr))
        return FALSE;
      deblockingOffset = 0;
#ifdef CISCO_G728_HACK
      if (codec == RTP_DataFrame::G728)
        swab((char *)deblockPtr, (char *)deblockPtr, frameSize);
#endif
    }
    else {
      if (!ReadFrame(bufferPtr))
        return FALSE;
#ifdef CISCO_G728_HACK
      if (codec == RTP_DataFrame::G728)
        swab((char *)bufferPtr, (char *)bufferPtr, frameSize);
#endif
      lastReadCount += frameSize;
      bufferPtr += frameSize;
      length -= frameSize;
    }
  }

  return TRUE;
}


BOOL OpalLineChannel::WriteFrame(const BYTE * buffer)
{
  if (device.WriteFrame(lineNumber, buffer))
    return TRUE;

#if PTRACING
  if (device.GetErrorNumber() != 0)
    PTRACE(1, "LID\tDevice write frame error: " << device.GetErrorText());
#endif
  return FALSE;
}


BOOL OpalLineChannel::Write(const void * buffer, PINDEX length)
{
  if (reading) {
    lastError = Miscellaneous;
    osError = EINVAL;
    return FALSE;
  }

  device.SetWriteFrameSize(lineNumber, length);
  PINDEX frameSize = device.GetWriteFrameSize(lineNumber);

  // Check for writing silence
  if (length == 0) {
    deblockingOffset = 0;  // Throw away any partial data
    lastWriteCount = 0;

    static const BYTE g723_sid_frame[24] = { 2 };
    static const BYTE g729_cng_frame[10] = { 0 }; // What is this?

    switch (codec) {
      case RTP_DataFrame::G7231 :
        return WriteFrame(g723_sid_frame);

      case RTP_DataFrame::G729 :
        return WriteFrame(g729_cng_frame);

      case RTP_DataFrame::PCMU:
      case RTP_DataFrame::PCMA:
        memset(deblockingBuffer.GetPointer(frameSize), 0x80, frameSize);
        break;

      default :
        memset(deblockingBuffer.GetPointer(frameSize), 0, frameSize);
        break;
    }
    return WriteFrame(deblockingBuffer);
  }

  /* The outer logical channel processing will automatically increment through
     the packet buffer by the value of lastWriteCount bytes. So even if length
     is set to 80 bytes for G.723.1 say, returning 24 will cause this function
     to be called again (in 30 milliseconds in the G.723.1 case) with len set
     to 56 and the buf pointer incremented 24 bytes.
   */
  const BYTE * bufferPtr = (const BYTE *)buffer;

  if (codec == RTP_DataFrame::G7231) {
    // Pick out special cases for G.723.1 based codec (variable length frames)
    switch (bufferPtr[0]&3) {
      case 0 : // 6.3kbps rate frame
        lastWriteCount = 24;
        break;
      case 1 : // 5.3kbps rate frame
        lastWriteCount = 20;
        break;
      case 2 : // a Silence Insertion Descriptor
        lastWriteCount = 4;
        break;
      case 3 : // repeat last CNG frame
        lastWriteCount = 1;
        break;
    }
    return WriteFrame(bufferPtr);
  }

  // Are reblocking output frame sizes to those expected by codec hardware.
  lastWriteCount = 0;

  while (length > 0) {
    // If have enough data and nothing in the reblocking buffer, just send it
    // straight on to the device.
    if (deblockingOffset == 0 && length >= frameSize) {
#ifdef CISCO_G728_HACK
      if (codec == RTP_DataFrame::G728)
        swab((char *)bufferPtr, (char *)bufferPtr, frameSize);
#endif
      if (!WriteFrame(bufferPtr))
        return FALSE;
      lastWriteCount += frameSize;
      bufferPtr += frameSize;
      length -= frameSize;
    }
    else {
      BYTE * savedFramePtr = deblockingBuffer.GetPointer(frameSize);

      // See if new chunk gives us enough for one frames worth
      if ((deblockingOffset + length) < frameSize) {
        // Nope, just copy bytes into buffer and return
        memcpy(savedFramePtr + deblockingOffset, bufferPtr, length);
        deblockingOffset += length;
        lastWriteCount += length;
        return TRUE;
      }

      /* Calculate bytes we want from the passed in buffer to fill a frame by
         subtracting from full frame width the amount we have so far. This also
         means the lastWriteCount is set to the correct amount of buffer we are
         grabbing this time around.
       */
      PINDEX left = frameSize - deblockingOffset;
      memcpy(savedFramePtr + deblockingOffset, bufferPtr, left);
      deblockingOffset = 0;

#ifdef CISCO_G728_HACK
      if (codec == RTP_DataFrame::G728)
        swab((char *)bufferPtr, (char *)bufferPtr, frameSize);
#endif

      // Write the saved frame out
      if (!WriteFrame(savedFramePtr))
        return FALSE;

      lastWriteCount += left;
      bufferPtr += left;
      length -= left;
    }
  }

  return TRUE;
}


///////////////////////////////////////////////////////////////////////////////

static struct {
  const char * description;

  unsigned bytesPerFrame;
  unsigned samplesPerFrame;
  unsigned rxFramesInPacket;
  unsigned txFramesInPacket;

  RTP_DataFrame::PayloadTypes   payloadType;
  H245_AudioCapability::Choices capabilitySubtype;

} CodecTypeInfo[] = {

  { "G.711-uLaw-64k{hw}", 8,    8, 240, 30, RTP_DataFrame::PCMU,  H245_AudioCapability::e_g711Ulaw64k },
  { "G.711-ALaw-64k{hw}", 8,    8, 240, 30, RTP_DataFrame::PCMA,  H245_AudioCapability::e_g711Alaw64k },
#ifdef CISCO_G728_HACK
  { "G.728{hw}",          20,  80,  24,  6, RTP_DataFrame::G728,  H245_AudioCapability::e_g728 },
#else
  { "G.728{hw}",          5,   20,  96, 20, RTP_DataFrame::G728,  H245_AudioCapability::e_g728 },
#endif
  { "G.729a{hw}",         10,  80,  24,  6, RTP_DataFrame::G729,  H245_AudioCapability::e_g729AnnexA },
  { "G.723.1{hw}",        24, 240,   8,  3, RTP_DataFrame::G7231, H245_AudioCapability::e_g7231 },
  { "GSM-0610{hw}",       33, 160,   7,  4, RTP_DataFrame::GSM,   H245_AudioCapability::e_gsmFullRate }

};


void H323_LIDCapability::AddAllCapabilities(const OpalLineInterfaceDevice & device,
                                            H323Capabilities & capabilities,
                                            PINDEX descriptorNum,
                                            PINDEX simultaneous)
{
  PIntArray codecsAvailable = device.GetPayloadTypes();
  for (PINDEX c = 0; c < codecsAvailable.GetSize(); c++) {
    H323_LIDCapability * cap = new H323_LIDCapability((RTP_DataFrame::PayloadTypes)codecsAvailable[c]);
    if (cap->IsValid() && !capabilities.FindCapability(cap->GetMainType(), cap->GetSubType()))
      capabilities.SetCapability(descriptorNum, simultaneous, cap);
    else
      delete cap;
  }
}


H323_LIDCapability::H323_LIDCapability(RTP_DataFrame::PayloadTypes rtpType)
  : H323AudioCapability(0, 0)
{
  codecTableIndex = 0;

  while (IsValid()) {
    if (CodecTypeInfo[codecTableIndex].payloadType == rtpType) {
      maxFrameSize = CodecTypeInfo[codecTableIndex].bytesPerFrame;
      rxFramesInPacket = CodecTypeInfo[codecTableIndex].rxFramesInPacket;
      txFramesInPacket = CodecTypeInfo[codecTableIndex].txFramesInPacket;
      break;
    }

    codecTableIndex++;
  }
}


BOOL H323_LIDCapability::IsValid() const
{
  return codecTableIndex < PARRAYSIZE(CodecTypeInfo);
}


PObject * H323_LIDCapability::Clone() const
{
  return new H323_LIDCapability(*this);
}


PString H323_LIDCapability::GetFormatName() const
{
  return CodecTypeInfo[codecTableIndex].description;
}


unsigned H323_LIDCapability::GetSubType() const
{
  return CodecTypeInfo[codecTableIndex].capabilitySubtype;
}


H323Codec * H323_LIDCapability::CreateCodec(H323Codec::Direction direction) const
{
  return new H323_LIDCodec(direction,
                           direction == H323Codec::Encoder ? txFramesInPacket : rxFramesInPacket,
                           codecTableIndex);
}


BOOL H323_LIDCapability::OnSendingPDU(H245_AudioCapability & pdu,
                                      unsigned packetSize) const
{
  pdu.SetTag(GetSubType());

  switch (pdu.GetTag()) {
    case H245_AudioCapability::e_gsmFullRate :
    {
      H245_GSMAudioCapability & gsm = pdu;
      gsm.m_audioUnitSize = packetSize*CodecTypeInfo[codecTableIndex].bytesPerFrame;
      break;
    }
    case H245_AudioCapability::e_g7231 :
    {
      H245_AudioCapability_g7231 & g7231 = pdu;
      g7231.m_maxAl_sduAudioFrames = packetSize;
      g7231.m_silenceSuppression = TRUE;
      break;
    }
    default :
    {
      PASN_Integer & value = pdu;
      value = packetSize;
    }
  }

  return TRUE;
}


BOOL H323_LIDCapability::OnReceivedPDU(const H245_AudioCapability & pdu,
                                       unsigned & packetSize)
{
  if (pdu.GetTag() != GetSubType())
    return FALSE;

  switch (pdu.GetTag()) {
    case H245_AudioCapability::e_gsmFullRate :
    {
      const H245_GSMAudioCapability & gsm = pdu;
      packetSize = gsm.m_audioUnitSize/CodecTypeInfo[codecTableIndex].bytesPerFrame;
      break;
    }
    case H245_AudioCapability::e_g7231 :
    {
      const H245_AudioCapability_g7231 & g7231 = pdu;
      packetSize = g7231.m_maxAl_sduAudioFrames;
      break;
    }
    default :
    {
      const PASN_Integer & value = pdu;
      packetSize = value;
    }
  }

  return TRUE;
}


/////////////////////////////////////////////////////////////////////////////

H323_LIDCodec::H323_LIDCodec(Direction direction, unsigned numFrames, unsigned index)
  : H323AudioCodec(direction, CodecTypeInfo[index].samplesPerFrame)
{
  rtpPayloadType = CodecTypeInfo[index].payloadType;
  codecTableIndex = index;
  packetSize = CodecTypeInfo[index].bytesPerFrame;

  /* Special case for G.711 encoder, note this helps with optimisation of
    sound data but will break if remote does not send the maximum number of
    bytes that it said it could. It is legal for the remote end to do so
    though no endpoints seem to actually do that.
   */
  if (packetSize == 8) {
    packetSize *= numFrames;
    samplesPerFrame *= numFrames;
  }

  PTRACE(3, "LID\tCreated codec: pt=" << rtpPayloadType
         << ", bytes=" << packetSize << ", samples=" << samplesPerFrame);
}


BOOL H323_LIDCodec::Open(H323Connection & connection)
{
  if (!H323AudioCodec::Open(connection))
    return FALSE;

  if (rawDataChannel->IsDescendant(OpalLineChannel::Class())) {
    OpalLineChannel * lineChannel = (OpalLineChannel*)rawDataChannel;
    if (lineChannel->SetCodec(rtpPayloadType))
      return TRUE;

    PTRACE(1, "LID\tCould not set codec " << rtpPayloadType);
    return FALSE;
  }

  PTRACE(1, "LID\tNo IxJ device set.");
  return FALSE;
}


BOOL H323_LIDCodec::Read(BYTE * buffer, unsigned & length, RTP_DataFrame &)
{
  // This reads to an H323_IxJChannel class
  if (!rawDataChannel->Read(buffer, packetSize))
    return FALSE;

  length = DetectSilence() ? 0 : rawDataChannel->GetLastReadCount();
  return TRUE;
}


BOOL H323_LIDCodec::Write(const BYTE * buffer,
                          unsigned length,
                          const RTP_DataFrame & /*frame*/,
                          unsigned & written)
{
  // This writes to an H323_IxJChannel class
  if (length > packetSize)
    length = packetSize;

  if (!rawDataChannel->Write(buffer, length))
    return FALSE;

  written = rawDataChannel->GetLastWriteCount();
  return TRUE;
}


unsigned H323_LIDCodec::GetBandwidth() const
{
  return CodecTypeInfo[codecTableIndex].bytesPerFrame*8/3;
}


unsigned H323_LIDCodec::GetAverageSignalLevel()
{
  return ((OpalLineChannel*)rawDataChannel)->GetDevice().GetAverageSignalLevel(0);
}


/////////////////////////////////////////////////////////////////////////////
