package CallFlowCanvas;

use strict;
use vars qw($VERSION @ISA @EXPORT @EXPORT_OK);

require Exporter;
use Tk;
use Carp;

use constant DEBUG                 => 0;
use constant DEBUG_GUI             => 0;
use constant DEBUG_SIP_MESSAGES    => 0;
use constant DEBUG_LOGICAL_NAMES   => 0;
use constant DEBUG_SNAPSHOTS       => 0;
use constant UEBER_DEBUG_SNAPSHOTS => 0;
use constant DEBUG_POSTSCRIPT      => 0;

@ISA = qw (Exporter AutoLoader);
@EXPORT = qw ( add_sip_message change_snapshot_preferences readjust_width );
@EXPORT_OK = qw ( get_call_id _construct_canvas );
$VERSION = '0.01';


# Preloaded methods go here.

# Autoload methods go after =cut, and are processed by the autosplit program.

1;
#__END__

#Protected methods and class variables
#!#!#!{
    my $batch_process = 0;
    #Side-effected by &$_draw_arrow
    my %settings = (arrow_start_vposition    => 15,
		    canvas_default_font      => '-Adobe-Courier-Medium-R-Normal--12-120-75-75-M-70-ISO8859-1',
		    break_font               => '-Adobe-Courier-Medium--Normal--24-120-75-75-M-70-ISO8859-1',
		    canvas_width             => 500, #minimum; adjusted below
		    canvas_height            => 400, #minimum; adjusted below
		    space_between_arrows     => 40,
		    server_start_hposition   => 50,
		    server_vposition         => 20,
		    space_between_servers    => 50,
		    server_box_hsize         => 170, #! Later, scale to text of name
		    server_box_vsize         => 50,
		    break_width              => 10,
		    break_offset_space       => 15
		    );
    
    my %colors =   (INVITE     => 'brown4', 
		    ACK        => 'green3',
		    OPTIONS    => 'orange',
		    BYE        => 'darkgoldenrod4',
		    REGISTER   => 'purple',

		    '1xx'      => 'black',
		    '200'      => 'forest green',
		    '3xx'      => 'blue',
		    '4xx'      => 'red',
		    '5xx'      => 'dark red',

		    sip_header_name    => 'black',
		    sip_header_value   => 'blue',
		    sdp_header_name    => 'dark slate grey',
		    sdp_header_value   => 'green',

		    canvas_background  => 'light yellow',
		    canvas_foreground  => 'black',
		    server_box_fill    => 'white',
		    server_box_outline => 'black',
		    
		    break              => 'black');

    #parameters that control the Postscript printing functionality
    #These are just defaults; if the user has a .tricorderrc, preferences
    #are loaded from there and passed in to the constructor, overriding
    #these values.
    my %default_snapshot_preferences = (file           => 'call_flow.eps',
					colormode      => 'mono',
					scaling_factor => 1,
					rotate         => 1, #landscape orientation
					page_length    => 1100, #pixels
					page_width     => 800, #pixels
					);

    my ($page_width, $page_length);

    #Protected methods
    my $_set_initial_positions =
    sub
    {
	my ($self) = @_;
	$self->{arrow_vposition} = $settings{server_box_vsize}
                                 + $settings{server_vposition}
                                 + $settings{arrow_start_vposition};
	$self->{arrow_number} = 1;
	$self->{server_box_hposition} = $settings{server_start_hposition};
    };

    #Apply the user's choice of heuristics to the recipient or sender of a
    #SIP message to guess whether it is coming from a server or just from
    #some random endpoint.
    my $_looks_like_a_server =
    sub
    {
	my ($self, $message, $endpoint) = @_;
	my $heuristic = $self->{heuristic} || '';

	if (ref $message ne 'Telephony::SipMessage')
	{
	    warn "\&\$_looks_like_a_server got an argument of type " . ref $message
		. "\n, which is not a SipMessage.  Skipping...";
	    return;
	}
    
	#default:  Treat everything as a server.
	return 1 if ($heuristic eq 'assume-server');

	#another option:  Discriminate between servers and endpoints based on
	#their IPs

	if ($heuristic eq 'use-vovida-ip-numbering-scheme')
	{
	    my $ip = $message->{"${endpoint}_ip"};
	    
	    #Check to see if its IP fits the Vovida SIP server numbering
	    #scheme.
	    $ip =~ m|\.(.+)$|m;  #Look at the
	    my $last_octet = $1; #last octet only.
	    if ($last_octet == 180 || #a UA marshal, by Vovida conventions
		$last_octet == 200 || #a redirect server, by Vovida
		                      #conventions
		$last_octet == 110 || #a 5300 marshal, by Vovida conventions
		$last_octet == 210)   #a 5300, by Vovida conventions
	    {
		return 1;
	    }
	    else
	    {
		return 0;
	    }
	}
	#another option:  Be really clever and look at the actual SIP message.
	elsif ($heuristic eq 'use-sip-hints')
	{
	    my $ip = $message->{"${endpoint}_ip"};
	    if ($message->{sip_uri} =~ /@(\.+?):/)
	    {
		#works for both hostnames and IPs...
		my $ip_from_URI = inet_ntoa (gethostbyname ($1));
		
		if ($endpoint eq 'destination')
		{
		    #Check to see if the IP from the SIP-URI matches the
		    #destination IP; if so, it's an endpoint, not a server.
		    if ($ip_from_URI eq $ip) { return 0; }
		    else { return 1; }
		}
		else #by deduction, ($endpoint eq $SOURCE)
		{
		    #Check to see if the host in the SIP-URI matches
		    #the From header;
		    #if so, it's an endpoint, not a server
		    my $ip_from_From = $message->get_sip_header ('From');
		    if ($ip_from_From eq $ip) { return 0; }
		    else { return 1; }
		}
	    }
	    else
	    {
		warn "Got a SIP message with an invalid SIP-URI:  \n"
		    . Data::Dumper->Dump ([ $message ]);
	    }
	}
	else
	{
	    #Default:  Assume it's a server, just to be safe.
	    return 1;
	}
    };

    my $_draw_server_box =
    sub
    {
	my ($self, $server) = @_;
	my $canvas = $self->{canvas};

	print "in \&\$_draw_server_box\n" if DEBUG;

	#! Later, make customizable through --naming-scheme option
	my $server_name = $server->{logical_name}
                       || $server->{ip} . ':' . $server->{port};
	print "server $server_name is\n" . Data::Dumper->Dump ([ $server ]) if DEBUG_SIP_MESSAGES;

	my $x1 = $self->{server_box_hposition};
	my $y1 = $settings{server_vposition};
	my $x2 = $self->{server_box_hposition} + $settings{server_box_hsize};
	my $y2 = $y1 + $settings{server_box_vsize};

	print "Drawing box from ($x1, $y1) to ($x2, $y2) (for $server_name)\n" if DEBUG;
	my $box = $canvas->createRectangle
	                    ($x1,
			     $y1,
			     $x2,
			     $y2,
			     fill    => $colors{server_box_fill},
			     outline => $colors{server_box_outline},
			     tags    => [ 'server_box' ]
			     );
    
	if (DEBUG)
	{
	    my @coords = $canvas->bbox (\$box);
	    print "Drew box with bounding coordinates " . @coords;
	}
	my $horizontal_midpoint = $x1 + $settings{server_box_hsize} / 2;
	my $vertical_midpoint   = $y1 + $settings{server_box_vsize} / 2;
	
	#Record the horizontal coordinate of the _center_ of the server box
	#so &$_draw_arrow can recall it later.
	$server->{hposition} = $horizontal_midpoint;

	$canvas->createText ($horizontal_midpoint,
			     $vertical_midpoint,
			     text    => $server_name,
			     #! width   => $settings{server_hsize},
			     fill    => $colors{server_box_outline},
			     font    => $settings{canvas_default_font},
			     justify => 'center',
			     tags    => 'server_label'
			     );

	my $bottom_of_the_canvas = $canvas->cget ('-height');
	$self->{server_box_hposition} += ($settings{server_box_hsize}
					  + $settings{space_between_servers});
    };

    my $_draw_lines_under_server_boxes =
    sub
    {
	my ($self, $bottom_of_the_canvas) = @_;
	my $canvas = $self->{canvas};
	map
	{
	    $canvas->createLine ($_->{hposition},
				 $settings{server_vposition}
				 + $settings{server_box_vsize},
				 $_->{hposition},
				 $bottom_of_the_canvas,
				 fill   => $colors{server_box_outline},
				 smooth => 1,
				 tags   => [ 'vertical_line' ]
				 );
	} @{$self->{server_list}};
    };

    #Resize the canvas to make it large enough for all that has been
    #drawn on it.
    my $_resize_canvas =
    sub
    {
	my ($self) = @_;
	my $canvas = $self->{canvas};
	my ($x1, $y1, $x2, $y2) = my @bounding_coordinates = $canvas->bbox ('all');
	print "Resizing canvas:  ($x1, $y1), ($x2, $y2)\n" if DEBUG;
	$canvas->configure (scrollregion => \@bounding_coordinates);
	&$_draw_lines_under_server_boxes ($self, $y2); #bottom of the canvas
    };

    my $_handle_new_servers =
    sub
    {
	my ($self, $message, $endpoint, $draw) = @_;

	#Add the source and destination servers if they do not already
	#exist.
	my $ip          = $message->{"${endpoint}_ip"} || '';
	my $port        = $message->{"${endpoint}_port"};
	my $ip_and_port = $ip . ':' . $port;
	print "Examining server $ip:$port\n" if DEBUG_SIP_MESSAGES;

	if ($ip &&
            ! exists ($self->{server_hash}->{$ip_and_port})
	    && &$_looks_like_a_server ($self, $message, $endpoint))
	{
	    my $logical_name = undef;
	    if ($self->{logical_names})
	    {
		$logical_name = $self->{server_map}->{$ip_and_port};
		print "Got logical name $logical_name for server $ip_and_port\n" if DEBUG_LOGICAL_NAMES;
	    }
	    my $server = Telephony::SipServer->new ($ip,
						    $port,
						    $logical_name);
	    $self->{server_hash}->{$ip_and_port} = $server;
	    push @{$self->{server_list}}, $server;
	    print "Found new server (IP $ip, port $port):\n"
		. Data::Dumper->Dump ([ $server ]) if DEBUG;
	    &$_draw_server_box ($self, $server);
	    &$_resize_canvas ($self);
	}
    };

    my $_draw_header_balloons =
    sub
    {
	my ($self, $message_number, $message, $x, $y) = @_;
	my $canvas = $self->{canvas};

	print "After call to \$&_draw_header_balloons, \$message_number is $message_number\n" if DEBUG;

	defined $message->{sip_headers}
        or return;

	my $sip_balloon = FancyBalloon->new ($message->{sip_headers},
					     $colors{sip_header_name},
					     $colors{sip_header_value},
					     $canvas);
	if (! defined $sip_balloon)
	{
	    warn "FancyBalloon->new returned undef; there is no SIP balloon to display.";
	    return;
	}

	print "Constructed a SIP FancyBalloon:\n" . Data::Dumper->Dump ([ $sip_balloon ]) if DEBUG;
    
	$sip_balloon->title ("$message_number. $message->{reqres}");
	#! $sip_balloon->place (-x => $x, -y => $y + 5); #in $self->{canvas},
	#its parent
	#! Ouch, that doesn't work.  I'll have to fiddle with the window
	#! manager!
	if (defined $message->{sdp_headers})
	{
	    my $sdp_balloon = FancyBalloon->new ($message->{sdp_headers},
						 $colors{sdp_header_name},
						 $colors{sdp_header_value},
						 $canvas);
	    
	    if (! defined $sdp_balloon)
	    {
		warn "FancyBalloon->new returned undef; there is no SDP balloon to display.";
		return;
	    }
	
	    print "Constructed an SDP FancyBalloon:" . Data::Dumper->Dump ([ $sdp_balloon ]) if DEBUG;
	    $sdp_balloon->title ("$message_number. $message->{reqres}");
	    #! $sdp_balloon->place (-x => $x,
	    #!                      -y => $y + 5 + $sip_balloon->height());
	    #in $canvas, its parent
	    #! Ouch, that doesn't work.  I'll have to fiddle with the window manager!
	}
    };

    my $_get_arrow_color =
    sub
    {
	my ($self) = @_;
	my $color;

	my $reqres = shift();
        print "Getting color" if (DEBUG || DEBUG_SIP_MESSAGES);
        print " for $reqres arrow..." if DEBUG_SIP_MESSAGES;
        print "\n" if (DEBUG || DEBUG_SIP_MESSAGES);

        #Response codes:  Match on the first digit (which indicates the
        #general type of response)
        if ($reqres =~ /^(\d)/)
        {
            $color = $colors{(grep /$1/, (keys %colors))[0]};
        }
        #Requests:  Match on the whole name
        else
        {
	    $color = $colors{$reqres};
        }
	
        #Set a default color for unknown reqres's.
        if (!defined $color)
        {
	    $color = 'black';
        }
        print "its color is $color\n" if DEBUG;
        return $color;
    };

    #! This function could be cleaned up considerably by making %server_hash
    #! into an object capable of returning false (i.e. '') when a value
    #! can't be looked up.
    my $_draw_arrow =
    sub
    {
	my ($self, $sip_message) = @_;
	unless (ref $sip_message eq 'Telephony::SipMessage')
	{
	    warn "\$sip_message is a " . ref ($sip_message);
	    return;
	}
	my $canvas = $self->{canvas};
	my $message_number = $self->{arrow_number};
	
	print "in \&\$_draw_arrow\n" if DEBUG;
	print "Drawing arrow for SIP message\n"
	    . Data::Dumper->Dump ([ $sip_message ])
	    if DEBUG_SIP_MESSAGES;
    
	my $source_ip   = $sip_message->{source_ip};
	my $source_port = $sip_message->{source_port};
	if (!defined $source_ip)
	{
	    warn "\&\$_draw_arrow got a message with no source IP; skipping...\n"
		. "\tthe message was " . Data::Dumper->Dump ([ $sip_message ]) . "\n\n" if DEBUG_SIP_MESSAGES;
	    return;
	}
	if (!defined $source_port)
	{
	    warn "\&\$_draw_arrow got a message with no source port; skipping...\n"
		. "\tthe message was " . Data::Dumper->Dump ([ $sip_message ]) . "\n\n" if DEBUG_SIP_MESSAGES;
	    return;
	}
	
	my $destination_ip   = $sip_message->{destination_ip};
	my $destination_port = $sip_message->{destination_port};
	if (!defined $destination_ip)
	{
	    warn "\&\$_draw_arrow got a message with no destination IP; skipping...\n"
	       . "\tthe message was " . Data::Dumper->Dump ([ $sip_message ]) . "\n\n" if DEBUG_SIP_MESSAGES;
	    return;
	}
	if (!defined $destination_port)
	{
	    warn "\&\$_draw_arrow got a message with no destination port; skipping...\n"
	       . "\tthe message was " . Data::Dumper->Dump ([ $sip_message ]) . "\n\n" if DEBUG_SIP_MESSAGES;
	    return;
	}

	my $source_ip_and_port      = $source_ip . ':' . $source_port;
	my $destination_ip_and_port = $destination_ip . ':' . $destination_port;
    
	my $source = $self->{server_hash}->{$source_ip_and_port};
	print "Got corresponding server:  " . Data::Dumper->Dump ([ $source ]) . "\n" if DEBUG_SIP_MESSAGES;
	my $source_hposition      = (ref $source eq 'Telephony::SipServer') ?
	                            $source->{hposition} :
				    0;

	my $destination = $self->{server_hash}->{$destination_ip_and_port};
	print "Got corresponding server:  " . Data::Dumper->Dump ([ $destination ]) . "\n" if DEBUG_SIP_MESSAGES;
	my $destination_hposition = (ref $destination eq 'Telephony::SipServer') ?    
	                            $destination->{hposition} :
				    0;    

	print "Drawing arrow for source $source_ip_and_port and destination $destination_ip_and_port...\n" if DEBUG_SIP_MESSAGES;

	my $x1 = $source_hposition;
	my $x2 = $destination_hposition;
	my $y1 = my $y2 = $self->{arrow_vposition};
	
	my $color_coding = &$_get_arrow_color ($sip_message->{reqres});
	print "Drawing $color_coding arrow from ($x1, $y1) to ($x2, $y2)\n" if DEBUG;
	$canvas->createLine ($x1,
			     $y1,
			     $x2,
			     $y2,
			     arrow  => 'last',
			     smooth => 1,
			     fill   => $color_coding,
			     tags => [ "message_arrow_$message_number" ]
			     );

	my $label = $canvas->createText ($x1 < $x2 ? $x1 : $x2,
					 $y1 - 12,
					 text    => "$message_number. $sip_message->{reqres}",
					 anchor => 'nw',
					 fill    => $color_coding,
					 font    => $settings{canvas_default_font},
					 tags    => [ "message_label_$message_number" ]
					 );
	$canvas->bind ($label,
		       '<Button-1>',
		       sub
		       {
			   print "Before call to \&\$draw_header_balloons, \$message_number is $message_number\n" if DEBUG;
			   &$_draw_header_balloons ($self,
						    $message_number,
						    $sip_message, #closure!
						    $x1 < $x2 ? $x1 : $x2,
						    $y1);
		       }
        );

        #Move down the screen...
        $self->{arrow_vposition} += $settings{space_between_arrows};
        $self->{arrow_number}++;
    };

    #Insert a number into a filename, so that the various pages written
    #by $self->_take_snapshot_page have unique names.
    my $_increment_filename =
    sub
    {
	my ($self) = @_;

	print "Incrementing $self->{filename} to " if DEBUG;
	
	print "(the number is $self->{file_number}; incrementing it to " if DEBUG;
	$self->{file_number}++;
	print "$self->{file_number})... " if DEBUG;
	#For Windows users' sake, put the number before the first dot,
	#if there is one.
	{   
	    my $filename    = $self->{filename};
	    my $file_number = $self->{file_number};
	    last if $filename =~ s|(\..*)|$file_number$1|;
	    $self->{filename} .= $self->{file_number};
	}
	print "$self->{filename}\n" if DEBUG;
    };

    my $_take_snapshot_page =
    sub
    {
	my ($self) = @_;
	my $canvas = $self->{canvas};
	#!my ($x1, $y1, $x2, $y2) = @{$self->{'x1', 'x2', 'y1', 'y2'}};
	my $x1 = $self->{x1};
	my $y1 = $self->{y1};
	my $x2 = $self->{x2};
	my $y2 = $self->{y2};

	print "Saving snapshot for coordinates ($x1, $y1), ($x2, $y2) in $self->{output_filename}\n" if DEBUG_SNAPSHOTS;
	my $width  = $x2 - $x1;
        my $height = $y2 - $y1;
        print "Width is $width; height is $height\n" if DEBUG_SNAPSHOTS == 2;

        my $snapshot_preferences = $self->{snapshot_preferences};
	print "Making postscript with " . Data::Dumper->Dump ([ $snapshot_preferences, '$snapshot_preferences' ]) if DEBUG_SNAPSHOTS;
        my $postscript = $canvas->postscript (x          => $x1,
                                              y          => $y1,
                                              width      => $width,
                                              height     => $height,
                                              colormode  => $snapshot_preferences->{colormode},
                                              rotate     => $snapshot_preferences->{rotate}
					      );

        print "Got postscript:\n$postscript" if DEBUG_POSTSCRIPT;
        my $output_filename = $self->{output_filename};
        if (! open DIAGRAM, ">$output_filename")
        {
            carp ("Could not open the Postscript output file $output_filename for writing:  $!");
            return;
        }
        print "Printing Postscript to $output_filename...\n" if DEBUG_SNAPSHOTS == 2;
        print DIAGRAM $postscript;
        if (! close DIAGRAM)
        {
            &show_error ("Could not close the Postscript output file $output_filename:  $!" );
            return;
        }
        print "Closed $output_filename.\n" if DEBUG_SNAPSHOTS == 2;
    };

    my $_do_snapshot =
    sub
    {
	my ($self) = @_;
	my $canvas = $self->{canvas};

	$canvas->createRectangle ($self->{x1},
				  $self->{y1},
				  $self->{x2},
				  $self->{y2},
				  outline => 'blue') if UEBER_DEBUG_SNAPSHOTS;
	&$_increment_filename ($self);
	&$_take_snapshot_page ($self);
    };

    my $_do_horizontal =
    sub
    {
	my ($self) = @_;

	if ($self->{x2} >= $self->{greatest_x})
	{
	    $self->{x2} = $self->{greatest_x};
	    &$_do_snapshot ($self);
	    return;
	}
	while (1) { &$_do_snapshot ($self); }
	continue
	{
	    $self->{x1} = $self->{x2};
	    $self->{x2} += $page_width;
	    if ($self->{x2} >= $self->{greatest_x})
	    {
		$self->{x2} = $self->{greatest_x};
		&$_do_snapshot ($self);
		return;
	    }
	}
    };

    my $_in_bbox =
    sub
    {
	my ($self, $arrow_id, $items) = @_;
	my @items = @$items;
	my $canvas = $self->{canvas};

	foreach my $item (@items)
	{
	    print "Looking in " . $canvas->gettags ($item) . "...\n" if DEBUG_SNAPSHOTS == 2;
	    return 1 if grep /$arrow_id/, $canvas->gettags ($item);
	}
	return 0;
    };

    #Push orphaned SIP messages onto the next page.  The idea
    #is to keep separate instances of a single request or response
    #together when they are propagating through a series of servers.
    #For instance, all the 200 messages in a response to a BYE should
    #appear on the same page, even if that means leaving some blank space
    #on the preceding page.
    my $_adjust_snapshot_boundaries =
    sub
    {
	my ($self) = @_;
	my $canvas = $self->{canvas};
	
	my $x1 = $self->{x1};
	my $y1 = $self->{y1};
	my $x2 = $self->{x2};
	my $y2 = $self->{y2};

	my @items = $canvas->find ('enclosed', $x1, $y1, $x2, $y2);

	print "The enclosed test gave coordinates ($items[0], $items[1]), ($items[3], $items[4])\n" if DEBUG_SNAPSHOTS;

	if (DEBUG_SNAPSHOTS == 2)
	{
	    print "Adjusting items:\n";
	    foreach my $item (@items)
	    {
		print "$item:  " . $canvas->gettags ($item) . "\n";
	    }
	    print "\n";
	}

	#The items are already ordered from first-drawn (i.e. lowest on the
	#canvas) to last-drawn (i.e. highest on the canvas).
 
	#If the last-drawn item is a label, it must be an orphan to a message
	#arrow on the following page.

	my @filtered_items;
	foreach my $item (@items)
	{
	    my $type = $canvas->type ($item);
	    print "$item is a $type.  " if DEBUG_SNAPSHOTS == 2;
	    if ($type ne 'text')
	    {
		#It's either an arrow or a bit of window dressing like a server
		#box's horizontal line.  Either way, it's guaranteed not to be an
		#orphan.  Include it.
		print "Keeping it...\n" if DEBUG_SNAPSHOTS == 2;
		push @filtered_items, $item;
	    }
	    else
	    {
		my $text = $canvas->itemcget ($item, '-text');
		print ":  '$text'.\n" if DEBUG_SNAPSHOTS == 2;
		#Keep server boxes and labels and the lines drawn under them
		if (grep /server|line/, $canvas->gettags ($item))
		{
		    push @filtered_items, $item;
		    print "It's a server glyph.  Keeping it.\n" if DEBUG_SNAPSHOTS == 2;
		    next;
		}
		$item =~ m|^(\d+)|;
		if (&$_in_bbox ($self, "message_arrow_$1", \@items))
		{
		    print "Keeping $item (a label with text $text)." if DEBUG_SNAPSHOTS == 2;
		    push @filtered_items, $item;
		}
		#Otherwise leave $item out of the new list, since its arrow is out.
		elsif (DEBUG_SNAPSHOTS == 2)
		{
		    print "Throwing out $item (a label with text $text):"
			. "\tIts corresponding message arrow was not included in this page.\n";
		}
	    }
	}

	#Find the type of SIP message label that belongs to the first
	#arrow on the next page (i.e. the first canvas item after the last
	#item on this page)
	my $last_item = $filtered_items[$#filtered_items];
	my $next_label = $canvas->find ('above', $last_item);
	until ($canvas->type ($next_label) eq 'text')
	{
	    $next_label = $canvas->find ('above', $next_label);
	    if (! defined $next_label)
	    {
		#An undefined label means we've reached the end of the output.
		@{$self->{'x1', 'y1', 'x2', 'y2'}} = $canvas->bbox (@items);
		last;
	    }
	    print "...and \$next_label is $next_label\n" if DEBUG_SNAPSHOTS == 2;
	}
	
	if (DEBUG_SNAPSHOTS == 2)
	{
	    my @tags = $canvas->gettags ($next_label);
	    print "The first label on the next page is "
		. Data::Dumper->Dump ([ $next_label ])
		. "\nIts tags are "
		. Data::Dumper->Dump ([ \@tags ]) if DEBUG_SNAPSHOTS == 2;
	}

	my $next_label_text = $canvas->itemcget ($next_label, '-text');
	print "\n\tIts text is $next_label_text\n" if DEBUG_SNAPSHOTS == 2;
	#Extract the SIP message type itself.
	my $next_SIP_message_type = '';
	if (defined $next_label_text)
	{
	    $next_label_text =~ m|\d+\.\s+(.*)|;
	    $next_SIP_message_type = $1;
	    print "Its message type is $next_SIP_message_type\n\n" if DEBUG_SNAPSHOTS == 2;
	}
	
	#Go backwards through the all the labels on the last page.
	for (my $i = $#filtered_items; $i > 0; $i--)
	{
	    my $label = $filtered_items[$i];
	    next unless $canvas->type ($label) eq 'text';
	    my $label_text = $canvas->itemcget ($label, '-text');
	    if (defined $label_text)
	    {
		print "Examining message '$label_text'...\n" if DEBUG_SNAPSHOTS == 2;
		$label_text =~ m|\d+\.\s+(.*)|;
		my $SIP_message_type = $1;
		if ($SIP_message_type eq $next_SIP_message_type)
		{
		    print "It's another $next_SIP_message_type.  Push it to the next page.\n" if DEBUG_SNAPSHOTS == 2;
		    #If it's the same kind of message, it belongs on the next
		    #page.  Remove both its label and its arrow from our list
		    #of items.
		    splice @filtered_items, $i, 2;
		}
		else
		{
		    print "It's a different type of message.\nLeave it where it is, and stop looking for orphans.\n\n" if DEBUG_SNAPSHOTS == 2;
		    last;
		}
	    }
	}

	#Finally, resize the page to include the new set of items, sans orphans.
	#(Don't mess with the x-coordinates; those must remain fixed to ensure
	#that the whole horizontal spread will be squeezed down to fit on one page.)
	($y1, $y2) = ($canvas->bbox (@filtered_items))[1,3];
	#Fudge the upper bound to include the first SIP message label.
	#Otherwise it will not quite fit (due to a quirk in the bbox method).
	$y1 -= 10 unless $self->{file_number} == 0; #Don't try it with the first snapshot!
    };

    #Cut the canvas down so there are no items (servers or messages)
    #hanging off its edges.
    my $_crop_canvas =
    sub
    {
	my ($self) = @_;
	my $canvas = $self->{canvas};

	#!my $x1 = \{$self->{x1}};
	#!my $x2 = \{$self->{x2}};
	#!my $y1 = \{$self->{y1}};
	#!my $y2 = \{$self->{y2}};

	print "Cropping from ($self->{x1}, $self->{y1}), ($self->{x2}, $self->{y2})...\n" if DEBUG_SNAPSHOTS == 2;
	my @items = $canvas->find ('enclosed', $self->{x1}, $self->{y1}, $self->{x2}, $self->{y2});
	unless (defined @items)
	{
	    print "Found no enclosed items!\n" if DEBUG_SNAPSHOTS;
	    return;
	}
    
        #Leave $x1 and $y1 intact no matter what.  They are guaranteed to be
	#justified just past the ending position of the last snapshot area
	#(or, in the case of the first snapshot, justified to the top of the page).
    
	#In the case of smart pagination, we know that the x range will include
	#the whole viewable area of the canvas, so we leave $x1 and $x2 alone too. 
        my $paginate_snapshots = $self->{paginate_snapshots};
	if ($paginate_snapshots eq 'smart')
	{
	    $self->{y2} = ($canvas->bbox (@items))[3];
	}
	#In the case of dumb pagination, the x range can vary.
	elsif ($paginate_snapshots eq 'dumb')
	{
	    my @bounding_coordinates = $canvas->bbox (@items);
	    $self->{x2} = $bounding_coordinates[2];
	    $self->{y2} = $bounding_coordinates[3];
	}
    
	print "         to   ($self->{x1}, $self->{y1}), ($self->{x2}, $self->{y2})\n\n" if DEBUG_SNAPSHOTS == 2;
	return \@items;
    };

#!#!#!}
#end of protected methods and class variables


#public methods

sub new
{
    my ($class, $parent, $canvas, $snapshot_preferences, $paginate_snapshots, $options, $sip_message) = @_;
    my %self;
    my $self = bless \%self, $class;
    $self{parent} = $parent;
    &$_set_initial_positions ($self);

    $self->{canvas} = $canvas;

    my %server_hash = (); $self->{server_hash} = \%server_hash;
    my @server_list = (); $self->{server_list} = \@server_list;
    my @sip_messages = []; $self->{sip_messages} = \@sip_messages;
    $self->add_sip_message ($sip_message) if ref $sip_message eq 'Telephony::SipMessage';

    if (defined $snapshot_preferences)
    {
	$self{snapshot_preferences} = $snapshot_preferences;
    }
    else
    {
	$self{snapshot_preferences} = \%default_snapshot_preferences;
    }
    $self{paginate_snapshots} = $paginate_snapshots;
    $self->{heuristic} = $options->{heuristic} if defined $options->{heuristic};
    $self->{logical_names} = $options->{logical_names} if defined $options->{logical_names};
    print "Created a new CallFlowCanvas:\n" . Data::Dumper->Dump ([ $self ]) if DEBUG;
    return $self;
}

sub _construct_canvas
{
    return $_[0]->Scrolled ('Canvas',
			    -scrollbars   => 'ose',
			    width         => $settings{canvas_width},
			    height        => $settings{canvas_height},
			    relief        => 'flat',
			    background    => $colors{canvas_background}
			    #foreground   => $colors{canvas_foreground}
			    #! font       => $settings{canvas_default_font}
			    );
}

sub hide
{
    my ($self) = @_;
    my $canvas = $self->{canvas};
    $canvas->delete ($canvas->find ('all'));
    &$_set_initial_positions ($self);
}

sub show
{
    my ($self) = @_;
    $self->restore_server_boxes();
    foreach my $message (@{$self->{sip_messages}})
    {
	&$_draw_arrow ($self, $message);# if ref $message eq 'Telephony::SipMessage';
    }
    &$_resize_canvas ($self);
}

sub add_sip_message
{
    my ($self, $message, $draw) = @_;
    my $this_canvas_call_id = $self->get_call_id();
    if (defined $this_canvas_call_id
	&& ($message->get_call_id() ne $this_canvas_call_id))
    {
	confess ('Got a SipMessage with the wrong call ID for this canvas!');
    }
    else #this canvas doesn't have a call ID defined yet.  Make one.
    {
	$self->{call_id} = $message->get_call_id();
    }
	
    &$_handle_new_servers ($self, $message, 'source', $draw);
    &$_handle_new_servers ($self, $message, 'destination', $draw);
    push @{$self->{sip_messages}}, $message;
    if ($draw)
    {
	&$_draw_arrow ($self, $message);
	&$_resize_canvas ($self);
    }
}

sub change_snapshot_preferences
{
    my ($self, $prefs) = @_;
    $self->{snapshot_preferences} = $prefs;
}

sub batch_mode
{
    if (@_)
    {
	($batch_process) = @_;
    }
    else
    {
	return $batch_process;
    }
}

sub get_call_id
{
    my ($self) = @_;
    return $self->{call_id};
}

#Make sure there is space for all the servers.
sub readjust_width
{
    my ($self) = @_;

    my $total_server_hspace =
    $self->{total_server_hspace} = @{$self->{server_list}}
                                   * ($settings{server_box_hsize}
				   + $settings{space_between_servers})
			           + $settings{server_start_hposition};
    if ($self->{canvas_width} < $total_server_hspace)
    {
	$self->{canvas_width} = $total_server_hspace;
    }
}

sub insert_break
{
    print "In CallFlowCanvas::insert_break...\n" if DEBUG;
    my ($self, $label) = @_;

    my $canvas = $self->{canvas};

    my ($leftmost_x, $rightmost_x) = ($canvas->bbox ('all'))[0, 2];
    print "Drawing line from $leftmost_x to $rightmost_x\n" if DEBUG;
    
    if (defined $label)
    {
	print "Drawing text \"$label\" at ($leftmost_x, " . $self->{arrow_vposition} + 2 . ")\n" if DEBUG;

	$canvas->createText ($leftmost_x,
			     $self->{arrow_vposition},
			     text    => $label,
			     font    => $settings{break_font},
			     #! width   => $settings{server_hsize},
			     fill    => $colors{break},
			     justify => 'right'
			     );
    }

    $self->{arrow_vposition} += $settings{break_offset_space};

    $canvas->createLine ($leftmost_x,
			 $self->{arrow_vposition},
			 $rightmost_x,
			 $self->{arrow_vposition},
			 width => $settings{break_width},
			 smooth => 1,
			 fill => $colors{break}
			 );
    &$_resize_canvas ($self);
    $self->{arrow_vposition} += $settings{break_width} = $settings{break_offset_space};
}

sub take_snapshot
{
    my ($self) = @_;
    my $canvas = $self->{canvas};

    #!my $x1 = \{; $self{x1}};
    #!my $y1 = \{; $self{y1}};
    #!my $x2 = \{; $self{x2}};
    #!my $y2 = \{; $self{y2}};

    $self->{output_filename} = $self->{snapshot_preferences}->{file} . $self->get_call_id();
    $self->{file_number} = 0;
    my ($least_x, $least_y, $greatest_x, $greatest_y) = $canvas->bbox ('all');
    if (DEBUG_SNAPSHOTS)
    {
	$canvas->createRectangle ($least_x, $least_y, $greatest_x, $greatest_y, outline => 'red');
    }

    print "Canvas dimensions are ($least_x, $least_y), ($greatest_x, $greatest_y)\n" if DEBUG_SNAPSHOTS;

    unless (defined $self->{paginate_snapshots}) #Just take one big snapshot.
    {
	#!@self->{'x1', 'y1', 'x2', 'y2'} = ($least_x, $least_y, $greatest_x, $greatest_y);
	$self->{x1} = $least_x;
	$self->{y1} = $least_y;
	$self->{x2} = $greatest_x;
	$self->{y2} = $greatest_y;

	&$_take_snapshot_page($self);
	return;
    }

    #Divide the canvas into pages of fixed width (configured through the
    #"Snapshot Properties" button).  If --paginate-snapshots was set to smart,
    #use additional formatting heuristics.

    if ($self->{paginate_snapshots} eq 'dumb')
    {
	$self->{x1} = $least_x;             #set here as well as in $self->_do_horizontal
	$self->{x2} = $self->{x1} + $page_width;    #so that $self->_crop_canvas will have all the
	                            #variables it needs to work the first
	                            #time it is invoked, before $self->_do_horizontal
	$self->{y1} = $least_y;
	$self->{y2} = $self->{y1} + $page_length;
	
	if ($self->{y2} >= $greatest_y)
	{
	    $self->{y2} = $greatest_y;
	    &$_do_horizontal ($self); #one last time...
	    return;
	}
	while (1)
	{
	    $self->_do_horizontal();
	}
	continue
	{
	    $self->{y1} = $self->{y2};
	    $self->{y2} += $page_length;
	    if ($self->{y2} >= $greatest_y)
	    {
		$self->{y2} = $greatest_y;
		&$_do_horizontal ($self);
		return;
	    }
	}
    }
    else # $paginate_snapshots eq 'smart', by deduction
    {
	$self->{x1} = $least_x;
	$self->{x2} = $greatest_x;
	$self->{y1} = $least_y;
	$self->{y2} = $self->{y1} + $page_length;
	if ($self->{y2} >= $greatest_y)
	{
	    &$_do_snapshot ($self); #one last time...
	    return;
	}
	while (1)
	{
	    &$_adjust_snapshot_boundaries ($self);
	    &$_do_snapshot ($self);
	}
	continue
	{
	    $self->{y1} = $self->{y2};
	    $self->{y2} += $page_length;
	    if ($self->{y2} >= $greatest_y)
	    {
		$self->{y2} = $greatest_y;
		&$_do_snapshot ($self);
		return;
	    }
	}
    }
}

sub last_sip_message
{
    my ($self) = @_;
    $self->take_snapshot() if $self->{batch_process};
}

#Erase servers and their window dressing from the canvas.
sub remove_server_boxes
{
    my ($self) = @_;
    my $canvas = $self->{canvas};
    $canvas->delete ($canvas->find ('withtag', 'server_box'));
    $canvas->delete ($canvas->find ('withtag', 'vertical_line'));
    $canvas->delete ($canvas->find ('withtag', 'server_label'));

}

#Restore servers and their window dressing to the canvas.
sub restore_server_boxes
{
    local $_;
    my ($self) = @_;
    map { &$_draw_server_box ($self, $_); } @{$self->{server_list}};    
}

=head1 NAME

CallFlowCanvas - Perl extension for blah blah blah

=head1 SYNOPSIS

  use CallFlowCanvas;
  blah blah blah

=head1 DESCRIPTION

Stub documentation for CallFlowCanvas was created by h2xs. It looks like the
author of the extension was negligent enough to leave the stub
unedited.

Blah blah blah.

=head1 AUTHOR

A. U. Thor, a.u.thor@a.galaxy.far.far.away

=head1 SEE ALSO

perl(1).

=cut
